# Copyright (c) 2023
# Manuel Cherep <mcherep@mit.edu>
# Nikhil Singh <nsingh1@mit.edu>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import jax
import jax.numpy as jnp
import chex
import evosax
from synthax.synth import BaseSynth
from typing import Callable


def polyphonic_synthapply(
    synth_applyfn: Callable,
    reshape_fn: Callable,
    n_synths: int,
    fparams: jnp.ndarray
) -> chex.Array:
    fparams_persynth = fparams.reshape(
        n_synths,
        fparams.shape[0],
        fparams.shape[1] // n_synths
    )
    params_persynth = reshape_fn(fparams_persynth)
    outputs = synth_applyfn(params_persynth)
    return outputs


class PolyphonicSynth:
    def __init__(
        self,
        synth: BaseSynth,
        n_synths: int,
        prng_key: jax.random.PRNGKey
    ):
        assert n_synths > 0, "Number of synthesizers must be greater than 0."

        self.synth = synth
        self.n_synths = n_synths
        self.params = [
            self.synth.init(jax.random.split(prng_key, n_synths)[i])
            for i in range(n_synths)
        ]

        self._init_params = jax.tree_map(lambda x : x[0], self.params[0])
        self.reshaper = evosax.ParameterReshaper(self._init_params)
        self._apply = jax.vmap(self.synth.apply)
        self.flatten = lambda paramlist : jnp.concatenate(
            [self.reshaper.flatten(p) for p in paramlist],
            axis=1
        )
        self.reshape = jax.vmap(self.reshaper.reshape)

        self.apply = lambda x : polyphonic_synthapply(
            self._apply,
            self.reshape,
            self.n_synths,
            x
        )

    @property
    def n_params(self):
        return self.reshaper.total_params * self.n_synths

    @property
    def fparams(self):
        return self.flatten(self.params)

    @property
    def sample_rate(self):
        return self.synth.sample_rate
