import jax.numpy as jnp
import chex
from synthax.parameter import ModuleParameterRange
from synthax.synth import BaseSynth
from synthax.modules.oscillators import SineVCO, SquareSawVCO, Noise
from synthax.modules.envelopes import ADSR
from synthax.modules.lfos import LFO
from synthax.modules.amplifiers import ControlRateVCA, VCA
from synthax.modules.mixers import AudioMixer, ModulationMixer
from synthax.modules.control import ControlRateUpsample
from synthax.modules.keyboard import MonophonicKeyboard


class VoiceSubsettable(BaseSynth):
    subset: str = "Basic"

    def setup(self):
        modules_spec = [
            ("keyboard", MonophonicKeyboard, {}),
            ("control_upsample", ControlRateUpsample, {}),
            ("vca", VCA, {})
        ]

        # Basic level: Single Oscillator, No LFOs, No Noise, Single ADSR
        if self.subset == "Basic":
            modules_spec.extend([
                ("adsr_1", ADSR, {}),
                ("lfo_1", LFO, {}),
                ("vco_1", SineVCO, {}),
                ("mod_matrix", ModulationMixer, {
                    "n_input": 2,
                    "n_output": 2
                })
            ])

        elif self.subset == "Noise":
            modules_spec.extend([
                ("adsr_1", ADSR, {}),
                ("lfo_1", LFO, {}),
                ("noise", Noise, {}),
                ("mod_matrix", ModulationMixer, {
                    "n_input": 2,
                    "n_output": 1
                })
            ])

        # Intermediate level: Dual Oscillators, No LFOs, No Noise
        elif self.subset == "Intermediate":
            modules_spec.extend([
                ("adsr_1", ADSR, {}),
                ("adsr_2", ADSR, {}),
                ("vco_1", SineVCO, {}),
                ("vco_2", SquareSawVCO, {}),
                ("mod_matrix", ModulationMixer, {
                    "n_input": 2,
                    "n_output": 4
                })
            ])

        # Advanced level: Dual Oscillators, Single LFO, No Noise
        elif self.subset == "Advanced":
            modules_spec.extend([
                ("adsr_1", ADSR, {}),
                ("adsr_2", ADSR, {}),
                ("lfo_1", LFO, {}),
                ("lfo_1_amp_adsr", ADSR, {}),
                ("lfo_1_rate_adsr", ADSR, {}),
                ("control_vca", ControlRateVCA, {}),
                ("control_upsample", ControlRateUpsample, {}),
                ("vco_1", SineVCO, {}),
                ("vco_2", SquareSawVCO, {}),
                ("mod_matrix", ModulationMixer, {
                    "n_input": 3,
                    "n_output": 4
                })
            ])

        # Full level: Dual Oscillators, Dual LFOs, Noise Generator
        elif self.subset == "Full":
            modules_spec.extend([
                ("adsr_1", ADSR, {}),
                ("adsr_2", ADSR, {}),
                ("lfo_1", LFO, {}),
                ("lfo_2", LFO, {}),
                ("lfo_1_amp_adsr", ADSR, {}),
                ("lfo_2_amp_adsr", ADSR, {}),
                ("lfo_1_rate_adsr", ADSR, {}),
                ("lfo_2_rate_adsr", ADSR, {}),
                ("control_vca", ControlRateVCA, {}),
                ("mod_matrix", ModulationMixer, {
                    "n_input": 4,
                    "n_output": 5
                }),
                ("vco_1", SineVCO, {}),
                ("vco_2", SquareSawVCO, {}),
                ("noise", Noise, {})
            ])

        # Common module for all levels
        modules_spec.append(("mixer", AudioMixer, {
            "n_input": {
                "Basic": 1,
                "Noise": 1,
                "Intermediate": 2,
                "Advanced": 2,
                "Full": 3
            }[self.subset],
            "level": ModuleParameterRange(
                minimum=0,
                maximum=1,
                curve={
                    "Basic": jnp.array([1]),
                    "Noise": jnp.array([1]),
                    "Intermediate": jnp.array([1, 1]),
                    "Advanced": jnp.array([1, 1]),
                    "Full": jnp.array([1, 1, 0.025])
                }[self.subset]
            )
        }))

        modules = {}
        for name, module, params in modules_spec:
            modules[name] = module(config=self.config, **params)

        self.modules = modules

    def __call__(self, *args, **kwargs) -> chex.Array:
        midi_f0, note_on_duration = self.modules["keyboard"]()

        signals = []

        adsr_1 = self.modules["adsr_1"](note_on_duration) if "adsr_1" in self.modules else None
        adsr_2 = self.modules["adsr_2"](note_on_duration) if "adsr_2" in self.modules else None

        if self.subset == "Basic":
            adsr_1 = self.modules["adsr_1"](note_on_duration)
            lfo_1 = self.modules["lfo_1"]()
            vco_1_amp, vco_1_pitch = self.modules["mod_matrix"](adsr_1, lfo_1)

            vco_1_out = self.modules["vca"](
                self.modules["vco_1"](midi_f0, self.modules["control_upsample"](vco_1_pitch)),
                self.modules["control_upsample"](vco_1_amp)
            )

            signals = [vco_1_out]
        elif self.subset == "Noise":
            adsr_1 = self.modules["adsr_1"](note_on_duration)
            lfo_1 = self.modules["lfo_1"]()
            (noise_amp,) = self.modules["mod_matrix"](adsr_1, lfo_1)

            noise_out = self.modules["vca"](
                self.modules["noise"](),
                self.modules["control_upsample"](noise_amp)
            )

            signals = [noise_out]
        elif self.subset == "Intermediate":
            adsr_1 = self.modules["adsr_1"](note_on_duration)
            adsr_2 = self.modules["adsr_2"](note_on_duration)

            vco_1_pitch, vco_1_amp, vco_2_pitch, vco_2_amp = self.modules["mod_matrix"](adsr_1, adsr_2)
            vco_1_out = self.modules["vca"](
                self.modules["vco_1"](midi_f0, self.modules["control_upsample"](vco_1_pitch)),
                self.modules["control_upsample"](vco_1_amp),
            )
            vco_2_out = self.modules["vca"](
                self.modules["vco_2"](midi_f0, self.modules["control_upsample"](vco_2_pitch)),
                self.modules["control_upsample"](vco_2_amp)
            )

            signals = [vco_1_out, vco_2_out]
        elif self.subset == "Advanced":
            lfo_1_rate = self.modules["lfo_1_rate_adsr"](note_on_duration)
            lfo_1_amp = self.modules["lfo_1_amp_adsr"](note_on_duration)
            lfo_1 = self.modules["control_vca"](self.modules["lfo_1"](lfo_1_rate), lfo_1_amp)

            vco_1_pitch, vco_1_amp, vco_2_pitch, vco_2_amp = self.modules["mod_matrix"](adsr_1, adsr_2, lfo_1)
            vco_1_out, vco_2_out = [self.modules["vca"](
                self.modules[vco](midi_f0, self.modules["control_upsample"](pitch)),
                self.modules["control_upsample"](amp),
            ) for vco, pitch, amp in [("vco_1", vco_1_pitch, vco_1_amp), ("vco_2", vco_2_pitch, vco_2_amp)]]

            signals = [vco_1_out, vco_2_out]
        elif self.subset == "Full":
            # ADSRs for modulating LFOs
            lfo_1_rate = self.modules["lfo_1_rate_adsr"](note_on_duration)
            lfo_2_rate = self.modules["lfo_2_rate_adsr"](note_on_duration)
            lfo_1_amp = self.modules["lfo_1_amp_adsr"](note_on_duration)
            lfo_2_amp = self.modules["lfo_2_amp_adsr"](note_on_duration)

            # Compute LFOs with envelopes
            lfo_1 = self.modules["control_vca"](self.modules["lfo_1"](lfo_1_rate), lfo_1_amp)
            lfo_2 = self.modules["control_vca"](self.modules["lfo_2"](lfo_2_rate), lfo_2_amp)

            # ADSRs for Oscillators and noise
            adsr_1 = self.modules["adsr_1"](note_on_duration)
            adsr_2 = self.modules["adsr_2"](note_on_duration)

            # Mix all modulation signals
            (vco_1_pitch,
            vco_1_amp,
            vco_2_pitch,
            vco_2_amp,
            noise_amp) = self.modules["mod_matrix"](adsr_1, adsr_2, lfo_1, lfo_2,)

            # Create signal and with modulations and mix together
            vco_1_out = self.modules["vca"](
                self.modules["vco_1"](midi_f0, self.modules["control_upsample"](vco_1_pitch)),
                self.modules["control_upsample"](vco_1_amp),
            )
            vco_2_out = self.modules["vca"](
                self.modules["vco_2"](midi_f0, self.modules["control_upsample"](vco_2_pitch)),
                self.modules["control_upsample"](vco_2_amp),
            )
            noise_out = self.modules["vca"](
                self.modules["noise"](),
                self.modules["control_upsample"](noise_amp)
            )

            signals = [vco_1_out, vco_2_out, noise_out]

        return self.modules["mixer"](*signals)
